/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.plugin.view;

import org.opengion.hayabusa.common.HybsSystem;
import org.opengion.hayabusa.common.HybsSystemException;
import org.opengion.hayabusa.db.DBTableModel;
import org.opengion.fukurou.util.StringUtil;
import org.opengion.hayabusa.html.TableFormatter;

import java.util.List;

/**
 * ヘッダ、フッタ、ボディを指定して作成する、自由レイアウトが可能な、カスタムテーブル表示クラスです。
 * このクラスは、ViewForm_HTMLFormatTable クラスの代替えとしても使用できます。
 * その場合は、thead のみ指定すれば、同じフォームが tbody にも適用されます。
 * これは、まさに、ViewForm_HTMLFormatTable と同じです。
 *
 * AbstractViewForm により、setter/getterメソッドのデフォルト実装を提供しています。
 * 各HTMLのタグに必要な setter/getterメソッドのみ，追加定義しています。
 *
 * AbstractViewForm を継承している為,ロケールに応じたラベルを出力させる事が出来ます。
 *
 * @og.group 画面表示
 *
 * @version  4.0
 * @author	 Kazuhiko Hasegawa
 * @since    JDK5.0,
 */
public class ViewForm_HTMLCustomTable extends ViewForm_HTMLTable	{
	//* このプログラムのVERSION文字列を設定します。	{@value} */
	private static final String VERSION = "4.0.0 (2005/08/31)" ;

	// 3.5.4.0 (2003/11/25) TableFormatter クラス追加
	// 4.3.1.0 (2008/09/08) protectedに変更
	protected TableFormatter 		headerFormat	= null;
	protected TableFormatter[]		bodyFormats		= null;
	protected TableFormatter		footerFormat	= null;
	protected int					bodyFormatsCount = 0;

	// 3.5.4.6 (2004/01/30) 初期値変更
	protected static final int BODYFORMAT_MAX_COUNT = 10;

	// 4.3.4.4 (2009/01/01)
//	/**
//	 * デフォルトコンストラクター
//	 *
//	 */
//	public ViewForm_HTMLCustomTable() {
//		super();
//	}

	/**
	 * DBTableModel から HTML文字列を作成して返します。
	 * startNo（表示開始位置）から、pageSize（表示件数）までのView文字列を作成します。
	 * 表示残りデータが pageSize 以下の場合は,残りのデータをすべて出力します。
	 *
	 * @og.rev 3.5.0.0 (2003/09/17) BODY要素の noClass 属性を追加。
	 * @og.rev 3.5.0.0 (2003/09/17) &lt;tr&gt;属性は、元のフォーマットのまま使用します。
	 * @og.rev 3.5.2.0 (2003/10/20) ヘッダー繰り返し属性（ headerSkipCount ）を採用
	 * @og.rev 3.5.3.1 (2003/10/31) skip属性を採用。headerLine のキャッシュクリア
	 * @og.rev 3.5.4.0 (2003/11/25) TableFormatter クラスを使用するように変更。
	 * @og.rev 3.5.5.0 (2004/03/12) systemFormat(例：[KEY.カラム名]形式等)の対応
	 * @og.rev 3.5.5.0 (2004/03/12) No 欄そのものの作成判断ロジックを追加
	 * @og.rev 3.5.5.7 (2004/05/10) [#カラム名] , [$カラム名] に対応
	 * @og.rev 3.5.6.0 (2004/06/18) '!' 値のみ 追加 既存の '$' は、レンデラー
	 * @og.rev 3.5.6.2 (2004/07/05) makeFormat 処理をgetTableHead メソッドから移動
	 * @og.rev 3.5.6.4 (2004/07/16) ヘッダーとボディー部をJavaScriptで分離
	 * @og.rev 4.0.0 (2005/01/31) 新規作成（getColumnClassName ⇒ getColumnDbType）
	 * @og.rev 3.7.0.3 (2005/03/01) getBgColorCycleClass に、選択行マーカーを採用
	 * @og.rev 4.3.1.0 (2008/09/08) フォーマットが設定されていない場合のエラー追加・編集行のみを表示する属性(isSkipNoEdit)追加
	 * @og.rev 4.3.3.0 (2008/10/01) noTransition属性対応
	 * @og.rev 4.3.7.4 (2009/07/01) tbodyタグの入れ子を解消(FireFox対応)
	 *
	 * @param  startNo	  表示開始位置
	 * @param  pageSize   表示件数
	 * @return	DBTableModel から作成された HTML文字列
	 */
	public String create( final int startNo, final int pageSize )  {
		if( getRowCount() == 0 ) { return ""; }	// 暫定処置

		// 4.3.1.0 (2008/09/08)
		if( headerFormat == null ) {
			String errMsg = "ViewTagで canUseFormat() = true の場合、Formatter は必須です。";
			throw new HybsSystemException( errMsg );
		}

		headerLine	 = null;		// 3.5.3.1 (2003/10/31) キャッシュクリア

		int lastNo = getLastNo( startNo, pageSize );
		int blc = getBackLinkCount();
		int hsc = getHeaderSkipCount();		// 3.5.2.0 (2003/10/20)
		int hscCnt = 1;						// 3.5.2.0 (2003/10/20)

		StringBuilder out = new StringBuilder( HybsSystem.BUFFER_LARGE );

		headerFormat.makeFormat( getDBTableModel() );	// 3.5.6.2 (2004/07/05) 移動

		out.append( getCountForm( startNo,pageSize ) );
		out.append( getHeader() );

		if( bodyFormatsCount == 0 ) {
			bodyFormats[0] = headerFormat ;
			bodyFormatsCount ++ ;
		}
		else {
			for( int i=0; i<bodyFormatsCount; i++ ) {
				bodyFormats[i].makeFormat( getDBTableModel() );
			}
		}

//		out.append("<tbody>").append( HybsSystem.CR ); // 4.3.7.4 (2009/07/01)
		int bgClrCnt = 0;
		for( int row=startNo; row<lastNo; row++ ) {
//			if( isSkip( row ) ) { continue; }		// 3.5.3.1 (2003/10/31)
			if( isSkip( row ) || isSkipNoEdit( row ) ) { continue; } // 4.3.1.0 (2008/09/08)
			for( int i=0; i<bodyFormatsCount; i++ ) {
				TableFormatter bodyFormat = bodyFormats[i];
				if( ! bodyFormat.isUse( row,getDBTableModel() ) ) { continue; }		// 3.5.4.0 (2003/11/25)
				out.append("<tbody").append( getBgColorCycleClass( bgClrCnt++,row ) );
				if( isNoTransition() ) { // 4.3.3.0 (2008/10/01)
					out.append( getHiddenRowValue( row ) );
				}
				out.append(">");	// 3.7.0.3 (2005/03/01)
				out.append( bodyFormat.getTrTag() );

				// 3.5.5.0 (2004/03/12) No 欄そのものの作成判断追加
				if( isNumberDisplay() ) {
					String ckboxTD = "<td" + bodyFormat.getRowspan() + ">";
					out.append( makeCheckbox( ckboxTD,row,blc ) );
				}

				int cl = 0;
				for( ; cl < bodyFormat.getLocationSize(); cl++ ) {
					String fmt = bodyFormat.getFormat(cl);
					int loc = bodyFormat.getLocation(cl);	// 3.5.5.0
					if( ! bodyFormat.isNoClass() && loc >= 0 ) {	// 3.5.5.7 (2004/05/10)
						StringBuilder newtg = new StringBuilder( HybsSystem.BUFFER_LARGE );
						newtg.append("<td class=\"");
						newtg.append( getColumnDbType(loc) );	// 4.0.0 (2005/01/31)
						newtg.append("\" ");
						String tdclass = newtg.toString();
						fmt = StringUtil.replace( bodyFormat.getFormat(cl) ,"<td", tdclass );
					}
					out.append( fmt );			// 3.5.0.0
					// 3.5.5.7 (2004/05/10) #,$ 対応
					if( loc >= 0 ) {
						switch( bodyFormat.getType(cl) ) {
							case '#' : out.append( getColumnLabel(loc) );		break;
							case '$' : out.append( getRendererValue(row,loc) );	break;
							case '!' : out.append( getValue(row,loc) );			break;
							default  : out.append( getValueLabel(row,loc) );	break;
						}
					}
					else {
						out.append( bodyFormat.getSystemFormat(row,loc) );
					}
				}
				out.append( bodyFormat.getFormat(cl) );
				out.append("</tbody>").append( HybsSystem.CR );
			}

		// 3.5.2.0 (2003/10/20) ヘッダー繰り返し属性（ headerSkipCount ）を採用
			if( hsc > 0 && hscCnt % hsc == 0 ) {
				out.append("<tbody class=\"row_h\"").append(" >");
				out.append( getHeadLine() );
				out.append("</tbody>");
				hscCnt = 1;
			}
			else {
				hscCnt ++ ;
			}
		}

		if( footerFormat != null ) {
			out.append( getTableFoot() );
		}

//		out.append("</tbody>").append( HybsSystem.CR ); // 4.3.7.4 (2009/07/01)
		out.append("</table>").append( HybsSystem.CR );

		out.append( getScrollBarEndDiv() );	// 3.8.0.3 (2005/07/15)
		return out.toString();
	}

	/**
	 * 内容をクリア（初期化）します。
	 *
	 * @og.rev 3.1.1.0 (2003/03/28) 同期メソッド（synchronized付き）を非同期に変更する。
	 * @og.rev 3.5.0.0 (2003/09/17) Noカラムに、表示を全て消せるように、class 属性を追加。
	 * @og.rev 3.5.4.0 (2003/11/25) TableFormatter クラスを使用するように変更。
	 *
	 */
	public void clear() {
		super.clear();
		headerFormat			= null;
		bodyFormats				= null;
		footerFormat			= null;
		bodyFormatsCount		= 0;
	}

	/**
	 * DBTableModel から テーブルのタグ文字列を作成して返します。
	 *
	 * @og.rev 3.5.0.0 (2003/09/17) &lt;tr&gt;属性は、元のフォーマットのまま使用します。
	 * @og.rev 3.5.1.0 (2003/10/03) Noカラムに、numberType 属性を追加
	 * @og.rev 3.5.2.0 (2003/10/20) ヘッダー繰り返し部をgetHeadLine()へ移動
	 * @og.rev 3.5.3.1 (2003/10/31) VERCHAR2 を VARCHAR2 に修正。
	 * @og.rev 3.5.4.0 (2003/11/25) TableFormatter クラスを使用するように変更。
	 * @og.rev 3.5.5.0 (2004/03/12) No 欄そのものの作成判断ロジックを追加
	 * @og.rev 3.5.6.2 (2004/07/05) makeFormat 処理をcreate メソッドの頭に移動
	 * @og.rev 3.5.6.5 (2004/08/09) thead に、id="header" を追加
	 * @og.rev 4.0.0 (2005/01/31) DBColumn の 属性(CLS_NM)から、DBTYPEに変更
	 *
	 * @return	テーブルのタグ文字列
	 */
	protected String getTableHead() {

		StringBuilder buf = new StringBuilder( HybsSystem.BUFFER_MIDDLE );
		// 3.5.5.0 (2004/03/12) No 欄そのものの作成判断追加
		if( isNumberDisplay() ) {
			buf.append("<colgroup class=\"X\" />");			// 4.0.0 (2005/01/31)
			buf.append("<colgroup class=\"BIT\" />");
			buf.append("<colgroup class=\"S9\" />");			// 4.0.0 (2005/01/31)
			buf.append(HybsSystem.CR);
		}

	// 3.5.2.0 (2003/10/20) ヘッダー繰り返し部をgetHeadLine()へ移動
		buf.append("<thead id=\"header\">").append( HybsSystem.CR );	// 3.5.6.5 (2004/08/09)
		buf.append( getHeadLine() );
		buf.append("</thead>").append( HybsSystem.CR );

		return buf.toString();
	}

	/**
	 * ヘッダー繰り返し部を、getTableHead()メソッドから分離。
	 *
	 * @og.rev 3.5.6.2 (2004/07/05) HTMLCustomTableScrollBar 用に新規作成
	 *
	 * @return	テーブルのタグ文字列
	 */
	protected String getHeadLine() {
		return getHeadLine( "<th" ) ;
	}

	/**
	 * ヘッダー繰り返し部を、getTableHead()メソッドから分離。
	 *
	 * @og.rev 3.5.2.0 (2003/10/20) 新規作成
	 * @og.rev 3.5.4.0 (2003/11/25) TableFormatter クラスを使用するように変更。
	 * @og.rev 3.5.4.3 (2004/01/05) useCheckControl 属性の機能を追加
	 * @og.rev 3.5.4.6 (2004/01/30) numberType="none" 時の処理を追加（Noラベルを出さない）
	 * @og.rev 3.5.4.7 (2004/02/06) ヘッダーにソート機能用のリンクを追加します。
	 * @og.rev 3.5.5.0 (2004/03/12) systemFormat(例：[KEY.カラム名]形式等)の対応
	 * @og.rev 3.5.5.0 (2004/03/12) No 欄そのものの作成判断ロジックを追加
	 * @og.rev 3.5.6.2 (2004/07/05) HTMLCustomTableScrollBar 用に引数追加
	 * @og.rev 3.7.0.1 (2005/01/31) 全件チェックコントロール処理変更
	 *
	 * @param	thTag タグの文字列
	 * @return	テーブルのタグ文字列
	 */
	protected String getHeadLine( final String thTag ) {
		if( headerLine != null ) { return headerLine; }		// キャッシュを返す。

		StringBuilder buf = new StringBuilder( HybsSystem.BUFFER_MIDDLE );

		buf.append( headerFormat.getTrTag() ).append( HybsSystem.CR );

		// 3.5.5.0 (2004/03/12) No 欄そのものの作成判断追加
		if( isNumberDisplay() ) {
			// 3.5.4.3 (2004/01/05) 追加分
			if( isUseCheckControl() && "checkbox".equals( getSelectedType() ) ) {
				buf.append( thTag ).append( headerFormat.getRowspan() ).append("></th>");
				buf.append( thTag ).append( headerFormat.getRowspan() );
				buf.append(">").append( getAllCheckControl() ).append( "</th>");
				buf.append( thTag ).append( headerFormat.getRowspan() );
				buf.append(">").append( getNumberHeader() ).append("</th>");	// 3.5.4.6 (2004/01/30)
			}
			else {
				buf.append( thTag ).append(" colspan=\"3\"");
				buf.append( headerFormat.getRowspan() );
				buf.append(">").append( getNumberHeader() ).append("</th>");	// 3.5.4.6 (2004/01/30)
			}
		}

		int cl = 0;
		for( ; cl < headerFormat.getLocationSize(); cl++ ) {
			buf.append( StringUtil.replace( headerFormat.getFormat(cl) ,"td","th" ));
			int loc = headerFormat.getLocation(cl);
			if( loc >= 0 ) { buf.append( getSortedColumnLabel(loc) ); }
		}
		buf.append( StringUtil.replace( headerFormat.getFormat(cl) ,"td","th" ) ).append( HybsSystem.CR );

		headerLine = buf.toString();
		return headerLine;
	}

	/**
	 * DBTableModel から テーブルのタグ文字列を作成して返します。
	 *
	 * @og.rev 3.5.0.0 (2003/09/17) &lt;tr&gt;属性は、元のフォーマットのまま使用します。
	 * @og.rev 3.5.1.0 (2003/10/03) Noカラムに、numberType 属性を追加
	 * @og.rev 3.5.4.0 (2003/11/25) TableFormatter クラスを使用するように変更。
	 * @og.rev 3.5.4.7 (2004/02/06) ヘッダーにソート機能用のリンクを追加します。
	 * @og.rev 3.5.5.0 (2004/03/12) systemFormat(例：[KEY.カラム名]形式等)の対応
	 * @og.rev 3.5.5.0 (2004/03/12) No 欄そのものの作成判断ロジックを追加
	 *
	 * @return	テーブルのタグ文字列
	 */
	protected String getTableFoot() {
		footerFormat.makeFormat( getDBTableModel() );

		StringBuilder buf = new StringBuilder( HybsSystem.BUFFER_MIDDLE );

		buf.append("<tfoot>").append( HybsSystem.CR );
		buf.append( footerFormat.getTrTag() ).append( HybsSystem.CR );

		// 3.5.5.0 (2004/03/12) No 欄そのものの作成判断追加
		if( isNumberDisplay() ) {
			buf.append(" <th");
			buf.append(" colspan=\"3\"");
			buf.append( footerFormat.getRowspan() );
			buf.append("></th>");
		}

		int cl = 0;
		for( ; cl < footerFormat.getLocationSize(); cl++ ) {
			int loc = footerFormat.getLocation(cl);
			if( loc >= 0 ) { buf.append( getSortedColumnLabel(loc) ); }
		}
		buf.append( footerFormat.getFormat(cl) ).append( HybsSystem.CR );
		buf.append("</tfoot>").append( HybsSystem.CR );

		return buf.toString();
	}

	/**
	 * フォーマットを設定します。
	 *
	 * @og.rev 3.5.4.0 (2003/11/25) 新規作成
	 * @og.rev 3.5.4.4 (2004/01/16) 配列の最大数を変更
	 * @og.rev 3.5.5.5 (2004/04/23) headerFormat が定義されていない場合はエラー
	 *
	 * @param list List
	 */
	public void setFormatterList( final List<TableFormatter> list ) {		// 4.3.3.6 (2008/11/15) Generics警告対応
		bodyFormats = new TableFormatter[BODYFORMAT_MAX_COUNT];

		bodyFormatsCount = 0;
		for( int i=0; i<list.size(); i++ ) {
			TableFormatter format = list.get( i );		// 4.3.3.6 (2008/11/15) Generics警告対応

			switch( format.getFormatType() ) {
				case TYPE_HEAD : headerFormat = format; break;
				case TYPE_BODY : bodyFormats[bodyFormatsCount++] = format; break;
				case TYPE_FOOT : footerFormat = format; break;
				default : String errMsg = "FormatterType の定義外の値が指定されました。";
				// 4.3.4.4 (2009/01/01)
						  throw new HybsSystemException( errMsg );
			}
		}

		// 3.5.5.5 (2004/04/23) headerFormat が定義されていない場合はエラー
		if( headerFormat == null ) {
			String errMsg = "h:thead タグの、フォーマットの指定は必須です。";
			throw new HybsSystemException( errMsg );
		}
	}

	/**
	 * フォーマットメソッドを使用できるかどうかを問い合わせます。
	 *
	 * @return  使用可能（true)/ 使用不可能 (false)
	 */
	public boolean canUseFormat() {
		return true;
	}

	/**
	 * ビューで表示したカラムの一覧をカンマ区切りで返します。
	 *
	 * @og.rev 5.1.6.0 (2010/05/01) 新規追加
	 *
	 * @return String ビューで表示したカラムの一覧
	 */
	public String getViewClms() {
		DBTableModel table = getDBTableModel();
		StringBuilder buf = new StringBuilder( HybsSystem.BUFFER_MIDDLE );
		for( int i=0; i<headerFormat.getLocationSize(); i++ ) {
			if( buf.length() > 0 ) { buf.append( ',' ); }
			buf.append( table.getColumnName( headerFormat.getLocation( i ) ) );
		}
		return buf.toString();
	}
}
